#define PERL_NO_GET_CONTEXT
#ifdef __cplusplus
extern "C" {
#endif
#include "EXTERN.h"
#include "perl.h"
#include "XSUB.h"
#ifdef __cplusplus
}
#endif
#include "sdblib/SDB.h"

#define NEED_sv_2pv_nolen
#include "ppport.h"

inline bool
bad_path(pTHX_ SV * path, I32 &numpath) {
   return ((!SvROK(path))
	   || (SvTYPE(SvRV(path)) != SVt_PVAV)
	   || ((numpath = av_len((AV *)SvRV(path))) < 0));
}

inline SV *
get_path(pTHX_ SDB *THIS, SV *path, I32 numpath) {
  THIS->reset();
  for (int n = 0; n < numpath; n++) {
    uint4 noid = SvIV(*av_fetch((AV *)SvRV(path), n, 0));
    THIS->set_next(noid);
  }
  return *av_fetch((AV *)SvRV(path), numpath, 0);
}

/*########################### INT4 ################################# */

typedef SV* (SV_CREATOR_NUM)(pTHX_ large_int num, const char* top);

inline SV*
creator_iv (pTHX_ large_int i, const char* top) {return newSViv(i);}
inline SV*
creator_nv (pTHX_ large_int i, const char* top) {return newSVnv(i);}
inline SV*
creator_pv (pTHX_ large_int i, const char* top) {
  const char *str = top + i;
  return newSVpv(str, strlen(str));
}

inline void
wrap_val(pTHX_ table < noid_type,noid_type >& tbl, std::pair<num_iter, num_iter> p, AV *results, SV_CREATOR_NUM key_creator,SV_CREATOR_NUM val_creator, bool pair = false) {
  const char* top = tbl.get_data_offset();
  for (num_iter cur = p.first; cur !=p.second; ++cur) {
    if(pair) av_push(results, key_creator(aTHX_ cur.get_key(), top));
    av_push(results, val_creator(aTHX_ cur.get_val(), top));
  }
}

inline void
map_int(pTHX_ table < noid_type,noid_type >& tbl,  large_int n, AV *results, SV_CREATOR_NUM key_creator,SV_CREATOR_NUM val_creator, bool pair = false) {
  wrap_val(aTHX_ tbl, num_iter::map_numeric_key(&tbl, n), results, key_creator, val_creator, pair);
}

inline void
map_str(pTHX_ table < noid_type,noid_type >& tbl, char* str, AV *results, SV_CREATOR_NUM key_creator,SV_CREATOR_NUM val_creator, bool pair = false, NodeConst::comparison_type tt = NodeConst::FULL) {
  wrap_val(aTHX_ tbl, num_iter::map_string_key(&tbl, str, tt), results, key_creator, val_creator, pair);
}

static SV_CREATOR_NUM*
choose_creator(NodeConst::TYPE type) {
  switch (type) {
  case NodeConst::INT4:
     return &creator_iv; break;
  case NodeConst::INT8:
    return &creator_nv; break;
  case NodeConst::STRING4:
    return &creator_pv; break;
  case NodeConst::CONTEXT4:
  case NodeConst::CONTEXT8:
    throw sdb_error_incomplete_path();
  default:
    fprintf(stderr, "can't find value creator for Type %d !\n", type);
    throw sdb_error("can't find value creator for Type");
  }
  return &creator_iv;
}

static large_int
convert_sv_to_num(pTHX_ SV* val, NodeConst::TYPE type) {
  switch (type) {
  case NodeConst::INT4:
     return SvIV(val); break;
  case NodeConst::INT8:
    return (large_int) SvNV(val); break;
  case NodeConst::STRING4:
  default:
    fprintf(stderr, "internal error: unknown type %d for convert_sv_to_num !\n", type);
    throw sdb_error("internal error: unknown type for convert_sv_to_num !");
  }
  return 0;
}

static void
map_common (pTHX_ table < noid_type,noid_type > tbl, SV* key, AV* results, bool is_pair = false, NodeConst::comparison_type tt = NodeConst::PREFIX) {
  NodeConst::TYPE keytype = tbl.get_keytype();
  NodeConst::TYPE valtype = tbl.get_valtype();
  SV_CREATOR_NUM* key_creator = choose_creator(keytype);
  SV_CREATOR_NUM* val_creator = choose_creator(valtype);
  switch (keytype) {
  case NodeConst::INT4:
  case NodeConst::INT8:
    map_int(aTHX_ tbl, convert_sv_to_num(aTHX_ key, keytype), results, *key_creator, *val_creator, is_pair);
    break;
  case NodeConst::STRING4:
    map_str(aTHX_ tbl, SvPV_nolen(key), results, *key_creator, *val_creator, is_pair, tt);
    break;
  default:
    fprintf(stderr, "Type %d is not supported in map_common!\n", keytype);
  }
}

static void
find_common (pTHX_ SDB *THIS, SV* key, AV* results, bool is_pair = false, NodeConst::comparison_type tt = NodeConst::FULL) {
  map_common(aTHX_ THIS->current()->get_find_map(), key, results, is_pair, tt);
}


/* ########################################################################## */

MODULE = Data::Array2ArrayMap::SDB		PACKAGE = Data::Array2ArrayMap::SDB

SDB *
SDB::new(name)
  char* name
CODE:
  try {
    RETVAL = new SDB(name);
  } catch (sdb_error e) {
    fprintf(stderr, "SDB::new -- %s\n", e.what());
    exit(-1);
  }
OUTPUT:
  RETVAL

void
SDB::DESTROY()

SV *
SDB::type(path)
  SV *path
INIT:
  I32 numpath = 0;
  if( bad_path(aTHX_ path, numpath) ) {
     XSRETURN_UNDEF;
   }
CODE:
   uint4 noid;
   SV* result =  newSVpv("__UNKNOWN__", 0);;
   try { 
     noid = SvIV(get_path(aTHX_ THIS, path,numpath));
     THIS->set_next(noid);
     result = newSVpv(NodeConst::get_type_name(THIS->current()->get_valtype()), 0);
     if (! NodeConst::is_valid_type(THIS->current()->get_valtype()))
       fprintf(stderr, "Unknown type %d!\n", THIS->current()->get_valtype());
   } catch (sdb_error_no_path e) {
   } catch (sdb_error e) {
     fprintf(stderr, "SDB::type -- %s\n", e.what());
   }
   RETVAL = result;
OUTPUT:
   RETVAL

SV *
SDB::get_ref(path)
  SV *path
INIT:
   I32 numpath = 0;
   if( bad_path(aTHX_ path, numpath) ) {
     XSRETURN_UNDEF;
   }
   AV * results = (AV *)sv_2mortal((SV *)newAV());
CODE:
   try { 
     SV * key = get_path(aTHX_ THIS, path, numpath);
     map_common(aTHX_ THIS->current()->get_get_map(), key, results);
   } catch (sdb_error_no_path e) {
   } catch (sdb_error e) {
     fprintf(stderr, "SDB::get_ref -- %s\n", e.what());
   }
   RETVAL = newRV_inc((SV *)results);
OUTPUT:
   RETVAL

SV *
SDB::find_ref(path)
  SV *path
INIT:
   I32 numpath = 0;
   if( bad_path(aTHX_ path, numpath) ) {
     XSRETURN_UNDEF;
   }
   AV * results = (AV *)sv_2mortal((SV *)newAV());
CODE:
   try { 
     SV * val = get_path(aTHX_ THIS, path, numpath);
     find_common(aTHX_ THIS, val, results, false);
   } catch (sdb_error_no_path e) {
   } catch (sdb_error e) {
     fprintf(stderr, "SDB::find_ref -- %s\n", e.what());
   }
   RETVAL = newRV_inc((SV *)results);
OUTPUT:
   RETVAL

SV *
SDB::find_pair(path)
  SV *path
INIT:
   I32 numpath = 0;
   if( bad_path(aTHX_ path, numpath) ) {
     XSRETURN_UNDEF;
   }
   AV * results = (AV *)sv_2mortal((SV *)newAV());
CODE:
   try { 
     SV * val = get_path(aTHX_ THIS, path, numpath);
     find_common(aTHX_ THIS, val, results, true);
   } catch (sdb_error_no_path e) {
   } catch (sdb_error e) {
     fprintf(stderr, "SDB::find_pair -- %s\n", e.what());
   }
   RETVAL = newRV_inc((SV *)results);
OUTPUT:
   RETVAL


SV *
SDB::find_pair_by_prefix(path)
  SV *path
INIT:
   I32 numpath = 0;
   if( bad_path(aTHX_ path, numpath) ) {
     XSRETURN_UNDEF;
   }
   AV * results = (AV *)sv_2mortal((SV *)newAV());
CODE:
   try { 
     SV * val = get_path(aTHX_ THIS, path, numpath);
     find_common(aTHX_ THIS, val, results, true, NodeConst::PREFIX);
   } catch (sdb_error_no_path e) {
   } catch (sdb_error e) {
     fprintf(stderr, "SDB::find_pair_by_prefix -- %s\n", e.what());
   }
   RETVAL = newRV_inc((SV *)results);
OUTPUT:
   RETVAL

int
SDB::exists_at(path)
   SV *path;
INIT:
   SV **top_noid;
   I32 numpath = 0;
   if( bad_path(aTHX_ path, numpath) ) {
     XSRETURN_UNDEF;
   }
CODE:
   top_noid = av_fetch ((AV*) SvRV(path), 0, 0);
   if (NULL == top_noid) RETVAL = 0;
   THIS->reset();
   if (THIS->exists_at(SvIV(*top_noid))) RETVAL = 1;
   else RETVAL=0;
OUTPUT:
   RETVAL


#   Local Variables:
#   mode: c++
#   End:
